/* Plan 9 support for GDB, using Plan 9's /proc.
   Copyright 1999 Free Software Foundation, Inc.
   Written by Michael Snyder at Cygnus Solutions.
   Based on work by Fred Fish, Stu Grossman, Geoff Noer, and others.

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "inferior.h"
#include "target.h"
#include "gdbcore.h"
#include "gdbcmd.h"
#include "gdbthread.h"

// #include <sys/errno.h>
#include <sys/wait.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>

#include <fcntl.h>	/* for O_RDONLY */
#include <unistd.h>	/* for "X_OK" */
#include "gdb_stat.h"	/* for struct stat */

static void procfs_open              PARAMS((char *, int));
static void procfs_attach            PARAMS ((char *, int));
static void procfs_detach            PARAMS ((char *, int));
static void procfs_resume            PARAMS ((int, int, enum target_signal));
static int  procfs_can_run           PARAMS ((void));	/* but can't hide */
static void procfs_stop              PARAMS ((void));
static void procfs_files_info        PARAMS ((struct target_ops *));
static void procfs_fetch_registers   PARAMS ((int));
static void procfs_store_registers   PARAMS ((int));
static void procfs_notice_signals    PARAMS ((int));
static void procfs_prepare_to_store  PARAMS ((void));
static void procfs_kill_inferior     PARAMS ((void));
static void procfs_mourn_inferior    PARAMS ((void));
static void procfs_create_inferior   PARAMS ((char *, char *, char **));
static int  procfs_wait              PARAMS ((int, 
					       struct target_waitstatus *));
static int  procfs_xfer_memory       PARAMS ((CORE_ADDR, 
					       char *, int, int, 
					       struct target_ops *));

static int  procfs_thread_alive      PARAMS ((int));

static void procfs_find_new_threads         PARAMS ((void));
static char *procfs_pid_to_str              PARAMS ((int));

struct target_ops procfs_ops;		/* the target vector */

static void
init_procfs_ops ()
{
  procfs_ops.to_shortname          = "plan9";
  procfs_ops.to_longname           = "Plan 9 /proc child process";
  procfs_ops.to_doc                = 
    "Plan 9 /proc child process (started by the \"run\" command).";
  procfs_ops.to_open               = procfs_open;
  procfs_ops.to_can_run            = procfs_can_run;
  procfs_ops.to_create_inferior    = procfs_create_inferior;
  procfs_ops.to_kill               = procfs_kill_inferior;
  procfs_ops.to_mourn_inferior     = procfs_mourn_inferior;
  procfs_ops.to_attach             = procfs_attach;
  procfs_ops.to_detach             = procfs_detach;
  procfs_ops.to_wait               = procfs_wait;
  procfs_ops.to_resume             = procfs_resume;
  procfs_ops.to_prepare_to_store   = procfs_prepare_to_store;
  procfs_ops.to_fetch_registers    = procfs_fetch_registers;
  procfs_ops.to_store_registers    = procfs_store_registers;
  procfs_ops.to_xfer_memory        = procfs_xfer_memory;
  procfs_ops.to_insert_breakpoint  =  memory_insert_breakpoint;
  procfs_ops.to_remove_breakpoint  =  memory_remove_breakpoint;
  procfs_ops.to_notice_signals     = procfs_notice_signals;
  procfs_ops.to_files_info         = procfs_files_info;
  procfs_ops.to_stop               = procfs_stop;

  procfs_ops.to_terminal_init      = terminal_init_inferior;
  procfs_ops.to_terminal_inferior  = terminal_inferior;
  procfs_ops.to_terminal_ours_for_output = terminal_ours_for_output;
  procfs_ops.to_terminal_ours      = terminal_ours;
  procfs_ops.to_terminal_info      = child_terminal_info;

  procfs_ops.to_find_new_threads   = procfs_find_new_threads;
  procfs_ops.to_thread_alive       = procfs_thread_alive;
  procfs_ops.to_pid_to_str         = procfs_pid_to_str;

  procfs_ops.to_has_all_memory    = 1;
  procfs_ops.to_has_memory        = 1;
  procfs_ops.to_has_execution      = 1;
  procfs_ops.to_has_stack          = 1;
  procfs_ops.to_has_registers      = 1;
  procfs_ops.to_stratum            = process_stratum;
  procfs_ops.to_has_thread_control = tc_schedlock;
  procfs_ops.to_magic              = OPS_MAGIC;
}

typedef struct gdb_gregset_t 
{
	unsigned	long	regs[128];
} gdb_gregset_t;

typedef struct gdb_fpregset_t 
{
	unsigned	long	regs[128];
} gdb_fpregset_t;

typedef struct procinfo {
  struct procinfo *next;
  int pid;			/* Process ID    */

  /* process state */
  int was_stopped;

  int ctl_fd;			/* File descriptor for /proc control file */
  int status_fd;		/* File descriptor for /proc status file */
  int mem_fd;			/* File descriptor for /proc mem file */
  int regs_fd;			/* File descriptor for /proc regs file */
  int fpregs_fd;			/* File descriptor for /proc fpregs file */

  gdb_gregset_t gregset;
  gdb_fpregset_t fpregset;

  int gregs_valid  : 1;
  int fpregs_valid : 1;
} procinfo;

/* Function prototypes for procinfo module: */

static procinfo *find_procinfo_or_die PARAMS ((int pid));
static procinfo *find_procinfo        PARAMS ((int pid));
static procinfo *create_procinfo      PARAMS ((int pid));
static void      destroy_procinfo     PARAMS ((procinfo *p));
static void      dead_procinfo        PARAMS ((procinfo *p, 
					       char *msg, int killp));
static int       open_procinfo_files  PARAMS ((procinfo *p));
static void      close_procinfo_files PARAMS ((procinfo *p));

/* The head of the procinfo list: */
static procinfo * procinfo_list;

static procinfo * 
find_procinfo (pid)
     int pid;
{
  procinfo *pi;

  for (pi = procinfo_list; pi; pi = pi->next)
    if (pi->pid == pid)
      break;

  return pi;
}

static procinfo *
find_procinfo_or_die (pid)
     int pid;
{
  procinfo *pi;

  pi = find_procinfo (pid);
  if (pi == NULL)
	error ("procfs: couldn't find pid %d in procinfo list.", pid);
  return pi;
}

static int
open_procinfo_files (pi)
     procinfo *pi;
{
	char tmp[80];

	sprintf(tmp, "/proc/%d/ctl", pi->pid);
	pi->ctl_fd = open(tmp, O_RDWR);
	if(pi->ctl_fd < 0)
		return 0;

	sprintf(tmp, "/proc/%d/status", pi->pid);
	pi->status_fd = open(tmp, O_RDONLY);
	if(pi->status_fd < 0)
		return 0;

	sprintf(tmp, "/proc/%d/mem", pi->pid);
	pi->mem_fd = open(tmp, O_RDWR);
	if(pi->mem_fd < 0)
		return 0;

	sprintf(tmp, "/proc/%d/regs", pi->pid);
	pi->regs_fd = open(tmp, O_RDWR);
	if(pi->regs_fd < 0)
		return 0;

	sprintf(tmp, "/proc/%d/fpregs", pi->pid);
	pi->fpregs_fd = open(tmp, O_RDWR);
	if(pi->fpregs_fd < 0)
		return 0;

	return 1;		/* success */
}

static procinfo *
create_procinfo (pid)
     int pid;
{
  procinfo *pi, *parent;

  if ((pi = find_procinfo (pid)))
    return pi;			/* Already exists, nothing to do. */

  pi = (procinfo *) xmalloc (sizeof (procinfo));
  memset (pi, 0, sizeof (procinfo));
  pi->pid = pid;
  pi->ctl_fd = pi->status_fd = pi->mem_fd = pi->regs_fd = pi->fpregs_fd = -1;

  /* Chain into list.  */
  pi->next = procinfo_list;
  procinfo_list = pi;
  return pi;
}

static void
close_procinfo_files (pi)
     procinfo *pi;
{
  if (pi->ctl_fd >= 0)
    close (pi->ctl_fd);
  if (pi->status_fd >= 0)
    close (pi->status_fd);
  if (pi->mem_fd >= 0)
    close (pi->mem_fd);
  if (pi->regs_fd >= 0)
    close (pi->regs_fd);
  if (pi->fpregs_fd >= 0)
    close (pi->fpregs_fd);
  pi->ctl_fd = pi->status_fd = pi->mem_fd = pi->regs_fd = pi->fpregs_fd = -1;
}

static void
destroy_procinfo (pi)
     procinfo  *pi;
{
  procinfo *ptr;

  /* Step one: unlink the procinfo from its list */
  if (pi == procinfo_list)
    procinfo_list = pi->next;
  else 
    for (ptr = procinfo_list; ptr; ptr = ptr->next)
      if (ptr->next == pi)
	{
	  ptr->next =  pi->next;
	  break;
	}

  /* Step two: close any open file descriptors */
  close_procinfo_files (pi);

  /* Step three: free the memory. */
  free (pi);

  /* Four: you catch the man! */
}

enum { NOKILL, KILL };

/*
 * Function: dead_procinfo
 *
 * To be called on a non_recoverable error for a procinfo.
 * Prints error messages, optionally sends a SIGKILL to the process,
 * then destroys the data structure.
 */

static void
dead_procinfo (pi, msg, kill_p)
     procinfo *pi;
     char     *msg;
     int       kill_p;
{
  char procfile[80];

  sprintf (procfile, "process %d", pi->pid);
  print_sys_errmsg (procfile, errno);
  if (kill_p == KILL)
    kill (pi->pid, SIGKILL);

  destroy_procinfo (pi);
  error (msg);
}

static int proc_stop_process         PARAMS ((procinfo *pi));
static int proc_set_gregs            PARAMS ((procinfo *pi));
static int proc_set_fpregs           PARAMS ((procinfo *pi));
static int proc_wait_for_stop        PARAMS ((procinfo *pi));
static int proc_run_process          PARAMS ((procinfo *pi, int step, int signo));

static gdb_gregset_t   *proc_get_gregs     PARAMS ((procinfo *pi));
static gdb_fpregset_t  *proc_get_fpregs    PARAMS ((procinfo *pi));

static void proc_warn  PARAMS ((procinfo *pi, char *func, int line));
static void proc_error PARAMS ((procinfo *pi, char *func, int line));

static void
proc_warn (pi, func, line)
     procinfo *pi;
     char     *func;
     int      line;
{
  char errmsg[128];
  sprintf (errmsg, "procfs: %s line %d, pid %d", func, line, pi->pid);
  print_sys_errmsg (errmsg, errno);
}

static void
proc_error (pi, func, line)
     procinfo *pi;
     char     *func;
     int      line;
{
  char errmsg[128];
  sprintf (errmsg, "procfs: %s line %d, pid %d", func, line, pi->pid);
  perror_with_name (errmsg);
}

static char *
proc_status(pi)
	procinfo *pi;
{
	char *p;
	static char buf[128];

	if(lseek(pi->status_fd, (off_t) 0, SEEK_SET) < 0)
		proc_error(pi, "proc_stopped", __LINE__);
	read(pi->status_fd, buf, sizeof(buf));
	p = buf+56+12;			/* Do better! */
	while(*p == ' ')
		p--;
	p[1] = '\0';
	return buf+56;			/* ditto */
}

static int
proc_stopped(pi)
	procinfo *pi;
{
	char *status;

	status = proc_status(pi);
	return strcmp(status, "Stopped") == 0 || strcmp(status, "Broken") == 0;
}

/*
 * Request the process to stop.  Not supposed to wait, but does.
 */

static int
proc_stop_process (pi)
     procinfo *pi;
{
	if(write (pi->ctl_fd, "stop", 4) != 4)
		return 0;

	return 1;
}

/*
 * Wait for the process to stop (block until it does).
 */

static int
proc_wait_for_stop (pi)
     procinfo *pi;
{
	if(write (pi->ctl_fd, "waitstop", 8) != 8)
		return 0;

	return 1;
}

/*
 * Function: proc_run_process
 *
 * Make the process or LWP runnable.
 * Options (not all are implemented):
 *   - single-step
 *   - clear current fault
 *   - clear current signal
 *   - abort the current system call
 *   - stop as soon as finished with system call
 *   - (ioctl): set traced signal set
 *   - (ioctl): set held   signal set
 *   - (ioctl): set traced fault  set
 *   - (ioctl): set start pc (vaddr)
 * Always clear the current fault.
 * Clear the current signal if 'signo' is zero.
 *
 * Arguments:
 *   pi		the process or LWP to operate on.
 *   step	if true, set the process or LWP to trap after one instr.
 *   signo	if zero, clear the current signal if any.
 *		if non-zero, set the current signal to this one.
 *
 * Returns non-zero for success, zero for failure. 
 */

static int
proc_run_process (pi, step, signo)
     procinfo *pi;
     int step;
     int signo;
{
  if (step)
    ;
  if (signo == 0)
    ;
  else if (signo != -1)		/* -1 means do nothing W.R.T. signals */
    ;

  return 1;
}

static gdb_gregset_t *
proc_get_gregs (pi)
     procinfo *pi;
{
	if(pi->gregs_valid)
		return &pi->gregset;	/* already got 'em */
	if(lseek(pi->regs_fd, (off_t) 0, SEEK_SET) < 0)
		return NULL;
	if(read(pi->regs_fd, &pi->gregset, sizeof(pi->gregset)) < 0)
		return NULL;
	pi->gregs_valid = 1;
	return &pi->gregset;
}

static int
proc_set_gregs (pi)
     procinfo *pi;
{
	gdb_gregset_t *gregs;

	if ((gregs = proc_get_gregs (pi)) == NULL)
		return 0;	/* get_regs has already warned */

	if (pi->regs_fd < 0 && open_procinfo_files (pi) == 0)
		return 0;
	/* do write here */
	/* Policy: writing the regs invalidates our cache. */
	pi->gregs_valid = 0;
	return 1;
}

static gdb_fpregset_t *
proc_get_fpregs (pi)
     procinfo *pi;
{
	if (pi->fpregs_valid)
		return &pi->fpregset;	/* already got 'em */
}

static int
proc_set_fpregs (pi)
     procinfo *pi;
{
	gdb_fpregset_t *fpregs;

	if ((fpregs = proc_get_fpregs (pi)) == NULL)
		return 0;		/* get_fpregs has already warned */

	if (pi->fpregs_fd < 0 && open_procinfo_files (pi) == 0)
		return 0;

	/* Policy: writing the regs invalidates our cache. */
	pi->fpregs_valid = 0;
	return 1;
}

/*
 *	I guess this stuff really belongs in i386-plan9-nat.c, or some such
 */

/* offsets into 386 Ureg to conform to ordering in i386-xdep.c */
static int regmap[] =
{
	7,	/* EAX */
	6,	/* ECX */
	5,	/* EDX */
	4,	/* EBX */
	17,	/* UESP */
	2,	/* EBP */
	1,	/* ESI */
	0,	/* EDI */
	14,	/* EIP */
	16,	/* EFL */
	15,	/* CS */
	18,	/* SS */
	11,	/* DS */
	10,	/* ES */
	9,	/* FS */
	8,	/* GS */
};

static void
supply_gregset(gregs)
	gdb_fpregset_t *gregs;
{
	int i;

	for(i = 0; i < sizeof(regmap)/sizeof(regmap[0]); i++)
		supply_register (i, (char *) &gregs->regs[regmap[i]]);
}

static int  do_attach PARAMS ((int pid));
static void do_detach PARAMS ((int signo));

/*
 * Function: procfs_debug_inferior
 *
 * Sets up the inferior to be debugged.
 * Registers to trace signals, hardware faults, and syscalls.
 * Note: does not set RLC flag: caller may want to customize that.
 *
 * Returns: zero for success (note! unlike most functions in this module)
 *   On failure, returns the LINE NUMBER where it failed!
 */

static int
procfs_debug_inferior (pi)
     procinfo *pi;
{
	/* open note file? wait file? */
	/* put proc in hang state for syscalls??? */

	return 0;
}

static void 
procfs_attach (args, from_tty)
     char *args;
     int from_tty;
{
  char *exec_file;
  int   pid;

  if (!args)
    error_no_arg ("process-id to attach");

  pid = atoi (args);
  if (pid == getpid ())
    error ("Attaching GDB to itself is not a good idea...");

  if (from_tty)
    {
      exec_file = get_exec_file (0);

      if (exec_file)
	printf_filtered ("Attaching to program `%s', %s\n", 
			 exec_file, target_pid_to_str (pid));
      else
	printf_filtered ("Attaching to %s\n", target_pid_to_str (pid));

      fflush (stdout);
    }
  inferior_pid = do_attach (pid);
  push_target (&procfs_ops);
}

static void 
procfs_detach (args, from_tty)
     char *args;
     int from_tty;
{
  char *exec_file;
  int   signo = 0;

  if (from_tty)
    {
      exec_file = get_exec_file (0);
      if (exec_file == 0)
	exec_file = "";
      printf_filtered ("Detaching from program: %s %s\n",
	      exec_file, target_pid_to_str (inferior_pid));
      fflush (stdout);
    }
  if (args)
    signo = atoi (args);
  
  do_detach (signo);
  inferior_pid = 0;
  unpush_target (&procfs_ops);		/* Pop out of handling an inferior */
}

static int
do_attach (pid)
     int pid;
{
  char errmsg[128];
  procinfo *pi;
  int fail;

  if ((pi = create_procinfo (pid)) == NULL)
    perror ("procfs: out of memory in 'attach'");

  if (!open_procinfo_files (pi))
    {
      fprintf_filtered (gdb_stderr, "procfs:%d -- ", __LINE__);
      sprintf (errmsg, "do_attach: couldn't open /proc file for process %d", 
	       pid);
      dead_procinfo (pi, errmsg, NOKILL);
    }

  /* Stop the process (if it isn't already stopped).  */
  if (proc_stopped(pi))
    {
      pi->was_stopped = 1;
      // proc_prettyprint_why (proc_why (pi), proc_what (pi), 1);
    }
  else
    {
      pi->was_stopped = 0;

      /* Now stop the process. */
      if (!proc_stop_process (pi))
	dead_procinfo (pi, "do_attach: couldn't stop the process.", NOKILL);
    }
  /* Save some of the /proc state to be restored if we detach.  */

  if ((fail = procfs_debug_inferior (pi)) != 0)
    dead_procinfo (pi, "do_attach: failed in procfs_debug_inferior", NOKILL);

  /* Let GDB know that the inferior was attached.  */
  attach_flag = 1;
  return pi->pid;
}

static void
do_detach (signo)
     int signo;
{
  procinfo *pi;

  pi = find_procinfo_or_die (inferior_pid);
  if (signo)
      proc_warn (pi, "do_detach, set_current_signal", __LINE__);

  attach_flag = 0;
  destroy_procinfo (pi);
}

/*
 * fetch_registers
 *
 * Since the /proc interface cannot give us individual registers,
 * we pay no attention to the (regno) argument, and just fetch them all.
 * This results in the possibility that we will do unnecessarily many
 * fetches, since we may be called repeatedly for individual registers.
 * So we cache the results, and mark the cache invalid when the process
 * is resumed.
 */

static void
procfs_fetch_registers (regno)
     int regno;
{
  gdb_fpregset_t *fpregs;
  gdb_gregset_t  *gregs;
  procinfo       *pi;

  pi  = find_procinfo_or_die (inferior_pid);

  if ((gregs = proc_get_gregs (pi)) == NULL)
    proc_error (pi, "fetch_registers, get_gregs", __LINE__);

  supply_gregset (gregs);

#if defined (FP0_REGNUM)	/* need floating point? */
  if ((regno >= 0 && regno < FP0_REGNUM) ||
      regno == PC_REGNUM  ||
#ifdef NPC_REGNUM
      regno == NPC_REGNUM ||
#endif
      regno == FP_REGNUM  ||
      regno == SP_REGNUM)
    return;			/* not a floating point register */

  if ((fpregs = proc_get_fpregs (pi)) == NULL)
    proc_error (pi, "fetch_registers, get_fpregs", __LINE__);

//  supply_fpregset (fpregs);
#endif
}

/* Get ready to modify the registers array.  On machines which store
   individual registers, this doesn't need to do anything.  On
   machines which store all the registers in one fell swoop, such as
   /proc, this makes sure that registers contains all the registers
   from the program being debugged.  */

static void
procfs_prepare_to_store ()
{
#ifdef CHILD_PREPARE_TO_STORE
  CHILD_PREPARE_TO_STORE ();
#endif
}

/*
 * store_registers
 *
 * Since the /proc interface will not read individual registers, 
 * we will cache these requests until the process is resumed, and
 * only then write them back to the inferior process.
 *
 * FIXME: is that a really bad idea?  Have to think about cases
 * where writing one register might affect the value of others, etc.
 */

static void
procfs_store_registers (regno)
     int regno;
{
  gdb_fpregset_t *fpregs;
  gdb_gregset_t  *gregs;
  procinfo       *pi;

   pi  = find_procinfo_or_die (inferior_pid);

  if ((gregs = proc_get_gregs (pi)) == NULL)
    proc_error (pi, "store_registers, get_gregs", __LINE__);

//  fill_gregset (gregs, regno);
  if (!proc_set_gregs (pi))
    proc_error (pi, "store_registers, set_gregs", __LINE__);

#if defined (FP0_REGNUM)	/* need floating point? */
  if ((regno >= 0 && regno < FP0_REGNUM) ||
      regno == PC_REGNUM  ||
#ifdef NPC_REGNUM
      regno == NPC_REGNUM ||
#endif
      regno == FP_REGNUM  ||
      regno == SP_REGNUM)
    return;			/* not a floating point register */

  if ((fpregs = proc_get_fpregs (pi)) == NULL)
    proc_error (pi, "store_registers, get_fpregs", __LINE__);

//  fill_fpregset (fpregs, regno);
  if (!proc_set_fpregs (pi))
    proc_error (pi, "store_registers, set_fpregs", __LINE__);
#endif
}

/*
 * Function: target_wait
 *
 * Retrieve the next stop event from the child process.
 * If child has not stopped yet, wait for it to stop.
 * Translate /proc eventcodes (or possibly wait eventcodes)
 * into gdb internal event codes.
 *
 * Return: id of process (and possibly thread) that incurred the event.
 *         event codes are returned thru a pointer parameter.
 */

static int  
procfs_wait (pid, status)
     int pid;
     struct target_waitstatus *status;
{
  procinfo *pi;
  int       temp, wstat, what;
  int       retval;
  int       retry = 0;

wait_again:

  retry++;
  wstat    = 0;
  what    = 0;
  retval   = -1;

  /* Find procinfo for main process */
  pi = find_procinfo_or_die (inferior_pid);
      /* We must assume that the status is stale now... */
//      pi->status_valid = 0;
      pi->gregs_valid  = 0;
      pi->fpregs_valid = 0;

      /* If child is not stopped, wait for it to stop.  */
      if (!proc_stopped (pi) && !proc_wait_for_stop (pi))
	{
	  /* wait_for_stop failed: has the child terminated? */
	  if (errno == ENOENT)
	    {
	      /* /proc file not found; presumably child has terminated. */
	      retval = wait (&wstat);	/* "wait" for the child's exit  */

	      if (retval != inferior_pid)	/* wrong child? */
		error ("procfs: couldn't stop process %d: wait returned %d\n",
		       inferior_pid, retval);
	      /* FIXME: might I not just use waitpid?
		 Or try find_procinfo to see if I know about this child? */
	    }
	  else
	    {
	      /* Unknown error from wait_for_stop. */
	      proc_error (pi, "target_wait (wait_for_stop)", __LINE__);
	    }
	}
      else
	{
//	      retval = pi->pid;
return inferior_pid;
	      /*
	       * If retval isn't in the threads database, add it.
	       */
	      if (retval > 0 &&
		  retval != inferior_pid &&
		  !in_thread_list (retval))
		{
		  /*
		   * We have a new thread.  
		   * We need to add it both to GDB's list and to our own.
		   * If we don't create a procinfo, resume may be unhappy 
		   * later.
		   */
		  printf_filtered ("[New %s]\n", target_pid_to_str (retval));
		  add_thread (retval);
		  if (find_procinfo (retval) == NULL)
		    create_procinfo (retval);
		}
	    }
//wstat = (what << 8) | 0177;
      if (status)
	store_waitstatus (status, wstat);

  return retval;
}

static int
procfs_xfer_memory (memaddr, myaddr, len, dowrite, target)
     CORE_ADDR memaddr;
     char *myaddr;
     int len;
     int dowrite;
     struct target_ops *target; /* ignored */
{
	procinfo *pi;
	int nbytes = 0;

	pi = find_procinfo_or_die (inferior_pid);
	if (pi->mem_fd < 0 && open_procinfo_files (pi) == 0) {
		proc_warn (pi, "xfer_memory, open_proc_files", __LINE__);
		return 0;
	}

	if (lseek (pi->mem_fd, (off_t) memaddr, SEEK_SET) != (off_t) memaddr)
		return 0;
	if (dowrite)
		nbytes = write (pi->mem_fd, myaddr, len);
	else
		nbytes = read (pi->mem_fd, myaddr, len);
	if (nbytes < 0)
		nbytes = 0;
	return nbytes;
}

/*
 * Function: invalidate_cache
 *
 * Called by target_resume before making child runnable.
 * Mark cached registers and status's invalid.
 * If there are "dirty" caches that need to be written back
 * to the child process, do that.
 *
 * File descriptors are also cached.  
 * As they are a limited resource, we cannot hold onto them indefinitely.
 * However, as they are expensive to open, we don't want to throw them
 * away indescriminately either.  As a compromise, we will keep the
 * file descriptors for the parent process, but discard any file
 * descriptors we may have accumulated for the threads.
 */


static void
invalidate_cache (pi)
     procinfo *pi;
{
  /*
   * About to run the child; invalidate caches and do any other cleanup.
   */

#if 0
  if (pi->gregs_dirty)
      if (!proc_set_gregs (pi))	/* flush gregs cache */
	proc_warn (pi, "target_resume, set_gregs",
		   __LINE__);
#ifdef FP0_REGNUM
  if (pi->fpregs_dirty)
       if (!proc_set_fpregs (pi))	/* flush fpregs cache */
	proc_warn (pi, "target_resume, set_fpregs", 
		   __LINE__);
#endif
#endif

  pi->gregs_valid   = 0;
  pi->fpregs_valid  = 0;
#if 0
  pi->gregs_dirty   = 0;
  pi->fpregs_dirty  = 0;
#endif
//  pi->status_valid  = 0;
}

/*
 * Function: target_resume
 *
 * Make the child process runnable.  Normally we will then call
 * procfs_wait and wait for it to stop again (unles gdb is async).
 *
 * Arguments:
 *  step:  if true, then arrange for the child to stop again 
 *         after executing a single instruction.
 *  signo: if zero, then cancel any pending signal.
 *         If non-zero, then arrange for the indicated signal 
 *         to be delivered to the child when it runs.
 *  pid:   if -1, then allow any child thread to run.
 *         if non-zero, then allow only the indicated thread to run.
 *******   (not implemented yet)
 */

static void
procfs_resume (pid, step, signo)
     int pid;
     int step;
     enum target_signal signo;
{
  procinfo *pi, *thread;
  int native_signo;

  pi = find_procinfo_or_die (inferior_pid);

  /* First cut: ignore pid argument */
  errno = 0;

  /* Convert signal to host numbering.  */
  if (signo == 0)
    native_signo = 0;
  else
    native_signo = target_signal_to_host (signo);

  /* Running the process voids all cached registers and status. */
  invalidate_cache (pi);

  if (pid != -1)
    {
    	warning ("procfs: no threads on Plan 9.");
      }

  if (!proc_run_process (pi, step, native_signo))
    {
      if (errno == EBUSY)
	warning ("resume: target already running.  Pretend to resume, and hope for the best!\n");
      else
	proc_error (pi, "target_resume", __LINE__);
    }
}

/*
 * Function: target_notice_signals
 *
 * Set up to trace signals in the child process.
 */

static void
procfs_notice_signals (pid)
     int pid;
{
  procinfo *pi = find_procinfo_or_die (pid);

    proc_error (pi, "notice_signals not implemented", __LINE__);
}

/*
 * Function: target_files_info
 *
 * Print status information about the child process.
 */

static void
procfs_files_info (ignore)
     struct target_ops *ignore;
{
  printf_filtered ("\tUsing the running image of %s %s via /proc.\n",
		   attach_flag? "attached": "child", 
		   target_pid_to_str (inferior_pid));
}

/*
 * Function: target_open
 *
 * A dummy: you don't open procfs.
 */

static void
procfs_open (args, from_tty)
     char *args;
     int from_tty;
{
  error ("Use the \"run\" command to start a Unix child process.");
}

/*
 * Function: target_can_run
 *
 * This tells GDB that this target vector can be invoked 
 * for "run" or "attach".
 */

static int
procfs_can_run ()
{
  return 1;
}

/*
 * Function: target_stop
 *
 * Stop the child process asynchronously, as when the
 * gdb user types control-c or presses a "stop" button.
 *
 * Works by sending kill(SIGINT) to the child's process group.
 */

static void
procfs_stop ()
{
  extern pid_t inferior_process_group;

  kill (-inferior_process_group, SIGINT);
}

/*
 * Function: unconditionally_kill_inferior
 *
 * Make it die.  Wait for it to die.  Clean up after it.
 * Note: this should only be applied to the real process, 
 * not to an LWP, because of the check for parent-process.
 * If we need this to work for an LWP, it needs some more logic.
 */

static void
unconditionally_kill_inferior (pi)
     procinfo *pi;
{
  if(write(pi->ctl_fd, "kill", 4) != 4)
    proc_warn (pi, "unconditionally_kill, cannot kill", __LINE__);
  destroy_procinfo (pi);

  /* If pi is GDB's child, wait for it to die. How? */
//  if (parent_pid == getpid ())
//       wait (NULL);
}

/*
 * Function: target_kill_inferior
 *
 * We're done debugging it, and we want it to go away.
 * Then we want GDB to forget all about it.
 */

static void 
procfs_kill_inferior ()
{
  if (inferior_pid != 0) /* ? */
    {
      procinfo *pi = find_procinfo (inferior_pid);

      if (pi)
	unconditionally_kill_inferior (pi);
      target_mourn_inferior ();
    }
}

/*
 * Function: target_mourn_inferior
 *
 * Forget we ever debugged this thing!
 */

static void 
procfs_mourn_inferior ()
{
  procinfo *pi;

  if (inferior_pid != 0)
    {
      pi = find_procinfo (inferior_pid);
      if (pi)
	destroy_procinfo (pi);
    }
  unpush_target (&procfs_ops);
  generic_mourn_inferior ();
}

/*
 * Function: init_inferior
 *
 * When GDB forks to create a runnable inferior process, 
 * this function is called on the parent side of the fork.
 * It's job is to do whatever is necessary to make the child
 * ready to be debugged, and then wait for the child to synchronize.
 */

static void 
procfs_init_inferior (pid)
     int pid;
{
  procinfo *pi;
  sigset_t signals;
  int fail;

  /* This routine called on the parent side (GDB side)
     after GDB forks the inferior.  */

  push_target (&procfs_ops);

  if ((pi = create_procinfo (pid)) == NULL)
    perror ("procfs: out of memory in 'init_inferior'");

  if (!open_procinfo_files (pi))
    proc_error (pi, "init_inferior, open_proc_files", __LINE__);

  /* If not stopped yet, wait for it to stop. */
  if (!proc_stopped(pi) && !(proc_wait_for_stop (pi)))
    dead_procinfo (pi, "init_inferior: wait_for_stop failed", KILL);

  /* Save some of the /proc state to be restored if we detach.  */
  /* FIXME: Why?  In case another debugger was debugging it?
     We're it's parent, for Ghu's sake! */

  if ((fail = procfs_debug_inferior (pi)) != 0)
    proc_error (pi, "init_inferior (procfs_debug_inferior)", fail);

  inferior_pid = pi->pid;

//#ifdef START_INFERIOR_TRAPS_EXPECTED
//  startup_inferior (START_INFERIOR_TRAPS_EXPECTED);
//#else
//  /* One trap to exec the shell, one to exec the program being debugged.  */
//  startup_inferior (2);
//#endif /* START_INFERIOR_TRAPS_EXPECTED */
}

/*
 * Function: set_exec_trap
 *
 * When GDB forks to create a new process, this function is called
 * on the child side of the fork before GDB exec's the user program.
 * Its job is to make the child minimally debuggable, so that the
 * parent GDB process can connect to the child and take over.
 * This function should do only the minimum to make that possible,
 * and to synchronize with the parent process.  The parent process
 * should take care of the details.
 */

static void
procfs_set_exec_trap ()
{
  /* This routine called on the child side (inferior side)
     after GDB forks the inferior.  It must use only local variables,
     because it may be sharing data space with its parent.  */

  procinfo *pi;

  if ((pi = create_procinfo (getpid ())) == NULL)
    perror_with_name ("procfs: create_procinfo failed in child.");

  if (open_procinfo_files (pi) == 0)
    {
      proc_warn (pi, "set_exec_trap, open_proc_files", __LINE__);
      gdb_flush (gdb_stderr);
      /* no need to call "dead_procinfo", because we're going to exit. */
      _exit (127);
    }

  /* Everyone else's (except OSF) method for tracing exec syscalls */
}

/*
 * Function: create_inferior
 *
 * This function is called BEFORE gdb forks the inferior process.
 * Its only real responsibility is to set things up for the fork, 
 * and tell GDB which two functions to call after the fork (one
 * for the parent, and one for the child).
 * 
 * This function does a complicated search for a unix shell program,
 * which it then uses to parse arguments and environment variables
 * to be sent to the child.  I wonder whether this code could not
 * be abstracted out and shared with other unix targets such as
 * infptrace?
 */

static void
procfs_create_inferior (exec_file, allargs, env)
     char *exec_file;
     char *allargs;
     char **env;
{
  char *shell_file = getenv ("SHELL");
  char *tryname;
  if (shell_file != NULL && strchr (shell_file, '/') == NULL)
    {

      /* We will be looking down the PATH to find shell_file.  If we
	 just do this the normal way (via execlp, which operates by
	 attempting an exec for each element of the PATH until it
	 finds one which succeeds), then there will be an exec for
	 each failed attempt, each of which will cause a PR_SYSEXIT
	 stop, and we won't know how to distinguish the PR_SYSEXIT's
	 for these failed execs with the ones for successful execs
	 (whether the exec has succeeded is stored at that time in the
	 carry bit or some such architecture-specific and
	 non-ABI-specified place).

	 So I can't think of anything better than to search the PATH
	 now.  This has several disadvantages: (1) There is a race
	 condition; if we find a file now and it is deleted before we
	 exec it, we lose, even if the deletion leaves a valid file
	 further down in the PATH, (2) there is no way to know exactly
	 what an executable (in the sense of "capable of being
	 exec'd") file is.  Using access() loses because it may lose
	 if the caller is the superuser; failing to use it loses if
	 there are ACLs or some such.  */

      char *p;
      char *p1;
      /* FIXME-maybe: might want "set path" command so user can change what
	 path is used from within GDB.  */
      char *path = getenv ("PATH");
      int len;
      struct stat statbuf;

      if (path == NULL)
	path = "/bin:/usr/bin";

      tryname = alloca (strlen (path) + strlen (shell_file) + 2);
      for (p = path; p != NULL; p = p1 ? p1 + 1: NULL)
	{
	  p1 = strchr (p, ':');
	  if (p1 != NULL)
	    len = p1 - p;
	  else
	    len = strlen (p);
	  strncpy (tryname, p, len);
	  tryname[len] = '\0';
	  strcat (tryname, "/");
	  strcat (tryname, shell_file);
	  if (access (tryname, X_OK) < 0)
	    continue;
	  if (stat (tryname, &statbuf) < 0)
	    continue;
	  if (!S_ISREG (statbuf.st_mode))
	    /* We certainly need to reject directories.  I'm not quite
	       as sure about FIFOs, sockets, etc., but I kind of doubt
	       that people want to exec() these things.  */
	    continue;
	  break;
	}
      if (p == NULL)
	/* Not found.  This must be an error rather than merely passing
	   the file to execlp(), because execlp() would try all the
	   exec()s, causing GDB to get confused.  */
	error ("procfs:%d -- Can't find shell %s in PATH",
	       __LINE__, shell_file);

      shell_file = tryname;
    }

//  fork_inferior (exec_file, allargs, env, procfs_set_exec_trap, 
//		 procfs_init_inferior, NULL, shell_file);

  /* We are at the first instruction we care about.  */
  /* Pedal to the metal... */

  proceed ((CORE_ADDR) -1, TARGET_SIGNAL_0, 0);
}

static void
procfs_find_new_threads ()
{
	/* No threads */
}

static int
procfs_thread_alive (pid)
     int pid;
{
  return 0;	/* No threads */
}

/*
 * Function: target_pid_to_str
 *
 * Return a string to be used to identify the thread in 
 * the "info threads" display.
 */

static char *
procfs_pid_to_str (pid)
     int pid;
{
	static char buf[80];

	sprintf (buf, "Process %d", pid);
	return &buf[0];
}

void
_initialize_procfs ()
{
  init_procfs_ops ();
  add_target (&procfs_ops);
}
