/* Read Plan9 symbol tables and convert to internal format, for GDB.
    Hacked together from dbxread.c.
    BUG: too much duplicated code with dbxread.c.

   Copyright 1986, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 1998
   Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* This module provides three functions: plan9_symfile_init,
   which initializes to read a symbol file; plan9_new_init, which 
   discards existing cached information when all symbols are being
   discarded; and plan9_symfile_read, which reads a symbol table
   from a file.
 */

#include "defs.h"
#include "gdb_string.h"

#if defined(USG) || defined(__CYGNUSCLIB__)
#include <sys/types.h>
#include <fcntl.h>
#endif

#include "obstack.h"
#include "gdb_stat.h"
#include <ctype.h>
#include "symtab.h"
#include "breakpoint.h"
#include "command.h"
#include "target.h"
#include "gdbcore.h"		/* for bfd stuff */
#include "libaout.h"		/* FIXME Secret internal BFD stuff for a.out */
#include "symfile.h"
#include "objfiles.h"
#include "buildsym.h"
#include "stabsread.h"
#include "gdb-stabs.h"
#include "demangle.h"
#include "language.h"		/* Needed inside partial-stab.h */
#include "complaints.h"

#include "aout/aout64.h"
#include "aout/stab_gnu.h"	/* We always use GNU stabs, not native, now */

struct symloc
  {
    /* Offset within the file symbol table of first local symbol for this
       file.  */

    int ldsymoff;

    /* Length (in bytes) of the section of the symbol table devoted to
       this file's symbols (actually, the section bracketed may contain
       more than just this file's symbols).  If ldsymlen is 0, the only
       reason for this thing's existence is the dependency list.  Nothing
       else will happen when it is read in.  */

    int ldsymlen;

    int symbol_offset;
  };

#define LDSYMOFF(p) (((struct symloc *)((p)->read_symtab_private))->ldsymoff)
#define LDSYMLEN(p) (((struct symloc *)((p)->read_symtab_private))->ldsymlen)
#define SYMBOL_OFFSET(p) (((struct symloc *)((p)->read_symtab_private))->symbol_offset)

/* Remember what we deduced to be the source language of this psymtab. */

static enum language psymtab_language = language_unknown;

/* Nonzero means give verbose info on gdb action.  From main.c.  */

extern int info_verbose;

/* This is the offset of the symbol table in the executable file. */

static unsigned symbol_table_offset;

/* The lowest text address we have yet encountered.  This is needed
   because in an a.out file, there is no header field which tells us
   what address the program is actually going to be loaded at, so we
   need to make guesses based on the symbols (which *are* relocated to
   reflect the address it will be loaded at).  */

static CORE_ADDR lowest_text_address;

/* Non-zero if there is any line number info in the objfile.  Prevents
   end_psymtab from discarding an otherwise empty psymtab.  */

static int has_line_numbers;

/* This is just to keep partial-stab.h happy */

static unsigned int file_string_table_offset;
static unsigned int next_file_string_table_offset;
static char *last_function_name;

/* Complaints about the symbols we have encountered are defined in dbxread.c.  */

extern struct complaint lbrac_complaint;
extern struct complaint string_table_offset_complaint;
extern struct complaint unknown_symtype_complaint;
extern struct complaint unknown_symchar_complaint;
extern struct complaint lbrac_rbrac_complaint;
extern struct complaint lbrac_unmatched_complaint;
extern struct complaint lbrac_mismatch_complaint;
extern struct complaint repeated_header_complaint;
extern struct complaint unclaimed_bincl_complaint;

/* During initial symbol readin, we need to have a structure to keep
   track of which psymtabs have which bincls in them.  This structure
   is used during readin to setup the list of dependencies within each
   partial symbol table. */

struct header_file_location
{
  char *name;			/* Name of header file */
  int instance;			/* See above */
  struct partial_symtab *pst;	/* Partial symtab that has the
				   BINCL/EINCL defs for this file */
};

/* The actual list and controling variables */
static struct header_file_location *bincl_list, *next_bincl;
static int bincls_allocated;

/* Local function prototypes */

static void
read_ofile_symtab PARAMS ((struct partial_symtab *));

static void
plan9_psymtab_to_symtab PARAMS ((struct partial_symtab *));

static void
plan9_psymtab_to_symtab_1 PARAMS ((struct partial_symtab *));

static void
read_plan9_symtab PARAMS ((struct objfile *));

static void
free_bincl_list PARAMS ((struct objfile *));

static struct partial_symtab *
  find_corresponding_bincl_psymtab PARAMS ((char *, int));

static void
add_bincl_to_list PARAMS ((struct partial_symtab *, char *, int));

static void
init_bincl_list PARAMS ((int, struct objfile *));

static char *
  plan9_next_symbol_text PARAMS ((struct objfile *));

static void
fill_symbuf PARAMS ((bfd *));

static void
record_minimal_symbol PARAMS ((char *, CORE_ADDR, int, struct objfile *));

static void
add_new_header_file PARAMS ((char *, int));

static void
add_old_header_file PARAMS ((char *, int));

static void
add_this_object_header_file PARAMS ((int));

static struct partial_symtab *
plan9_start_psymtab PARAMS ((struct objfile *, char *, CORE_ADDR, int,
		       struct partial_symbol **, struct partial_symbol **));

static struct partial_symtab *
plan9_end_psymtab PARAMS ((struct partial_symtab *pst, char **include_list, int num_includes,  int capping_symbol_offset,
	CORE_ADDR capping_text, struct partial_symtab **dependency_list, int number_dependencies, int textlow_not_set));

static void
plan9_process_one_symbol PARAMS ((int type, int desc, CORE_ADDR valu, char *name, struct section_offsets *section_offsets, struct objfile *objfile));

/* Free up old header file tables */

static void
free_header_files ()
{
  if (this_object_header_files)
    {
      free ((PTR) this_object_header_files);
      this_object_header_files = NULL;
    }
  n_allocated_this_object_header_files = 0;
}

/* Allocate new header file tables */

static void
init_header_files ()
{
  n_allocated_this_object_header_files = 10;
  this_object_header_files = (int *) xmalloc (10 * sizeof (int));
}

/* Add header file number I for this object file
   at the next successive FILENUM.  */

static void
add_this_object_header_file (i)
     int i;
{
  if (n_this_object_header_files == n_allocated_this_object_header_files)
    {
      n_allocated_this_object_header_files *= 2;
      this_object_header_files
	= (int *) xrealloc ((char *) this_object_header_files,
		       n_allocated_this_object_header_files * sizeof (int));
    }

  this_object_header_files[n_this_object_header_files++] = i;
}

/* Add to this file an "old" header file, one already seen in
   a previous object file.  NAME is the header file's name.
   INSTANCE is its instance code, to select among multiple
   symbol tables for the same header file.  */

static void
add_old_header_file (name, instance)
     char *name;
     int instance;
{
  register struct header_file *p = HEADER_FILES (current_objfile);
  register int i;

  for (i = 0; i < N_HEADER_FILES (current_objfile); i++)
    if (STREQ (p[i].name, name) && instance == p[i].instance)
      {
	add_this_object_header_file (i);
	return;
      }
  complain (&repeated_header_complaint, name, symnum);
}

/* Add to this file a "new" header file: definitions for its types follow.
   NAME is the header file's name.
   Most often this happens only once for each distinct header file,
   but not necessarily.  If it happens more than once, INSTANCE has
   a different value each time, and references to the header file
   use INSTANCE values to select among them.

   dbx output contains "begin" and "end" markers for each new header file,
   but at this level we just need to know which files there have been;
   so we record the file when its "begin" is seen and ignore the "end".  */

static void
add_new_header_file (name, instance)
     char *name;
     int instance;
{
  register int i;
  register struct header_file *hfile;

  /* Make sure there is room for one more header file.  */

  i = N_ALLOCATED_HEADER_FILES (current_objfile);

  if (N_HEADER_FILES (current_objfile) == i)
    {
      if (i == 0)
	{
	  N_ALLOCATED_HEADER_FILES (current_objfile) = 10;
	  HEADER_FILES (current_objfile) = (struct header_file *)
	    xmalloc (10 * sizeof (struct header_file));
	}
      else
	{
	  i *= 2;
	  N_ALLOCATED_HEADER_FILES (current_objfile) = i;
	  HEADER_FILES (current_objfile) = (struct header_file *)
	    xrealloc ((char *) HEADER_FILES (current_objfile),
		      (i * sizeof (struct header_file)));
	}
    }

  /* Create an entry for this header file.  */

  i = N_HEADER_FILES (current_objfile)++;
  hfile = HEADER_FILES (current_objfile) + i;
  hfile->name = savestring (name, strlen (name));
  hfile->instance = instance;
  hfile->length = 10;
  hfile->vector
    = (struct type **) xmalloc (10 * sizeof (struct type *));
  memset (hfile->vector, 0, 10 * sizeof (struct type *));

  add_this_object_header_file (i);
}

static void
plan9_symfile_read (objfile, mainline)
     struct objfile *objfile;
     int mainline;
{
	struct cleanup *back_to;

	init_minimal_symbol_collection ();
	back_to = make_cleanup ((make_cleanup_func) discard_minimal_symbols, 0);

	/* Read stabs data from executable file and define symbols. */
	read_plan9_symtab (objfile);

	install_minimal_symbols (objfile);
	do_cleanups (back_to);
}

/* Initialize anything that needs initializing when a completely new
   symbol file is specified.  */

static void
plan9_new_init (ignore)
     struct objfile *ignore;
{
  stabsread_new_init ();
  buildsym_new_init ();
  init_header_files ();
}

static void
plan9_symfile_init (objfile)
     struct objfile *objfile;
{
	bfd *sym_bfd;

	sym_bfd = objfile->obfd;

	/* Allocate struct to keep track of the symfile */
	objfile->sym_stab_info = (struct dbx_symfile_info *)
		xmmalloc (objfile->md, sizeof (struct dbx_symfile_info));
	memset ((PTR) objfile->sym_stab_info, 0, sizeof (struct dbx_symfile_info));

	DBX_TEXT_SECTION (objfile) = bfd_get_section_by_name (sym_bfd, ".text");
	DBX_DATA_SECTION (objfile) = bfd_get_section_by_name (sym_bfd, ".data");
	DBX_BSS_SECTION (objfile) = bfd_get_section_by_name (sym_bfd, ".bss");
}

/* Perform any local cleanups required when we are done with a particular
   objfile.  I.E, we are in the process of discarding all symbol information
   for an objfile, freeing up all memory held for it, and unlinking the
   objfile struct from the global list of known objfiles. */

static void
plan9_symfile_finish (objfile)
     struct objfile *objfile;
{
  if (objfile->sym_stab_info != NULL)
    {
      if (HEADER_FILES (objfile) != NULL)
	{
	  register int i = N_HEADER_FILES (objfile);
	  register struct header_file *hfiles = HEADER_FILES (objfile);

	  while (--i >= 0)
	    {
	      free (hfiles[i].name);
	      free (hfiles[i].vector);
	    }
	  free ((PTR) hfiles);
	}
      mfree (objfile->md, objfile->sym_stab_info);
    }
	free_header_files ();
}

/* dbx allows the text of a symbol name to be continued into the
   next symbol name!  When such a continuation is encountered
   (a \ at the end of the text of a name)
   call this function to get the continuation.  */

static char *
plan9_next_symbol_text (objfile)
     struct objfile *objfile;
{
#if 0
  struct internal_nlist nlist;

  if (symbuf_idx == symbuf_end)
    fill_symbuf (symfile_bfd);

  symnum++;
  INTERNALIZE_SYMBOL (nlist, &symbuf[symbuf_idx], symfile_bfd);
  OBJSTAT (objfile, n_stabs++);

  symbuf_idx++;

  return nlist.n_strx + stringtab_global + file_string_table_offset;
#endif
fprintf(stderr, "you lose\n");
abort();
return 0;
}

 /* Initialize the list of bincls to contain none and have some
   allocated.  */

static void
init_bincl_list (number, objfile)
     int number;
     struct objfile *objfile;
{
  bincls_allocated = number;
  next_bincl = bincl_list = (struct header_file_location *)
    xmmalloc (objfile->md, bincls_allocated * sizeof (struct header_file_location));
}

/* Add a bincl to the list.  */

static void
add_bincl_to_list (pst, name, instance)
     struct partial_symtab *pst;
     char *name;
     int instance;
{
  if (next_bincl >= bincl_list + bincls_allocated)
    {
      int offset = next_bincl - bincl_list;
      bincls_allocated *= 2;
      bincl_list = (struct header_file_location *)
	xmrealloc (pst->objfile->md, (char *) bincl_list,
		   bincls_allocated * sizeof (struct header_file_location));
      next_bincl = bincl_list + offset;
    }
  next_bincl->pst = pst;
  next_bincl->instance = instance;
  next_bincl++->name = name;
}

/* Given a name, value pair, find the corresponding
   bincl in the list.  Return the partial symtab associated
   with that header_file_location.  */

static struct partial_symtab *
find_corresponding_bincl_psymtab (name, instance)
     char *name;
     int instance;
{
  struct header_file_location *bincl;

  for (bincl = bincl_list; bincl < next_bincl; bincl++)
    if (bincl->instance == instance
	&& STREQ (name, bincl->name))
      return bincl->pst;

  complain (&repeated_header_complaint, name, symnum);
  return (struct partial_symtab *) 0;
}

/* Free the storage allocated for the bincl list.  */

static void
free_bincl_list (objfile)
     struct objfile *objfile;
{
  mfree (objfile->md, (PTR) bincl_list);
  bincls_allocated = 0;
}

static void
read_plan9_symtab (objfile)
     struct objfile *objfile;
{
	bfd *sym_bfd;
	CORE_ADDR val;
	int n, type, desc;
	unsigned char *syms, *p, *ep;
	char *name, buf[5];
	enum minimal_symbol_type ms_type;
	int section, index;
	asection *bfd_section;
	CORE_ADDR text_addr;
	int text_size;
	int sym_off;
	int symnum;

	/* Some definitions cribbed from dbxread */
	register char *namestring;
	int nsl;
	int past_first_source_file = 0;
	CORE_ADDR last_o_file_start = 0;
	CORE_ADDR last_function_start = 0;
	struct cleanup *back_to;
	int textlow_not_set;

	/* Current partial symtab */
	struct partial_symtab *pst;

	/* List of current psymtab's include files */
	char **psymtab_include_list;
	int includes_allocated;
	int includes_used;

	/* Index within current psymtab dependency list */
	struct partial_symtab **dependency_list;
	int dependencies_used, dependencies_allocated;

	sym_bfd = objfile->obfd;

	/* FIXME POKING INSIDE BFD DATA STRUCTURES */
	sym_off = sym_bfd->origin + obj_sym_filepos (sym_bfd);
	n = exec_hdr (sym_bfd)->a_syms;

	if(n == 0)
		return;

	val = bfd_seek (sym_bfd, sym_off, SEEK_SET);
	if (val < 0)
		perror_with_name (objfile->name);

	symbol_table_offset = sym_off;

	pst = (struct partial_symtab *) 0;

	includes_allocated = 30;
	includes_used = 0;
	psymtab_include_list = (char **) alloca (includes_allocated *
					   sizeof (char *));

	dependencies_allocated = 30;
	dependencies_used = 0;
	dependency_list =
		(struct partial_symtab **) alloca (dependencies_allocated *
				       sizeof (struct partial_symtab *));

	/* Init bincl list */
	init_bincl_list (20, objfile);
	back_to = make_cleanup ((make_cleanup_func) free_bincl_list, objfile);

	last_source_file = NULL;

	syms = (unsigned char *)bfd_malloc(n);
	if (syms == NULL)
		perror_with_name (objfile->name);
	if ((int)bfd_read (syms, 1, n, sym_bfd) != n) {
		free (syms);
		perror_with_name (objfile->name);
	}

 	p = syms;
	ep = p+n;
	for(;;) {
		if(p >= ep)
			break;
		val = bfd_h_get_32(sym_bfd, p);
		name = (char*)&p[5];
		switch(p[4] & ~0x80) {
		case 'T':
		case 'L':
			ms_type = mst_text;
			section = SECT_OFF_TEXT;
			bfd_section = DBX_TEXT_SECTION (objfile);
			break;
		case 't':
		case 'l':
			ms_type = mst_file_text;
			section = SECT_OFF_TEXT;
			bfd_section = DBX_TEXT_SECTION (objfile);
			break;
		case 'D':
			ms_type = mst_data;
			section = SECT_OFF_DATA;
			bfd_section = DBX_DATA_SECTION (objfile);
			break;
		case 'd':
			ms_type = mst_file_data;
			section = SECT_OFF_DATA;
			bfd_section = DBX_DATA_SECTION (objfile);
			break;
		case 'B':
			ms_type = mst_bss;
			section = SECT_OFF_BSS;
			bfd_section = DBX_BSS_SECTION (objfile);
			break;
		case 'b':
			ms_type = mst_file_bss;
			section = SECT_OFF_BSS;
			bfd_section = DBX_BSS_SECTION (objfile);
			break;
		case 'z':
			p += 5+1;
			for(;;) {
				if(p+1 >= ep)
					break;
				index = (p[0]<<8) | p[1];
				p += 2;
				if(index == 0)
					break;
			}
			continue;
		case 'X':			/* encapsulated stabs */
			buf[0] = name[0];
			buf[1] = name[1];
			buf[2] = '\0';
			type = strtoul(buf, (char **)NULL, 16);
			buf[0] = name[4];
			buf[1] = name[5];
			buf[2] = name[6];
			buf[3] = name[7];
			buf[4] = '\0';
			desc = strtoul(buf, (char **)NULL, 16);
			symnum = p-syms;
			p += 5;
			p += strlen(p)+1;

/*Now throw up.  */
#define SET_NAMESTRING()	namestring = name+8;
#define CUR_SYMBOL_TYPE type
#define CUR_SYMBOL_VALUE val
#define CUR_SYMBOL_STRX 0
#define symbol_size 1
// #define DBXREAD_ONLY
#define START_PSYMTAB(ofile,fname,low,symoff,global_syms,static_syms)\
  plan9_start_psymtab(ofile, fname, low, symoff, global_syms, static_syms)
#define END_PSYMTAB(pst,ilist,ninc,c_off,c_text,dep_list,n_deps,textlow_not_set)\
  plan9_end_psymtab(pst,ilist,ninc,c_off,c_text,dep_list,n_deps,textlow_not_set)

#include "partial-stab.h"
			continue;
		default:
			fprintf(stderr, "ignored symbol type `%c'\n", p[4] & ~0x80);
		case 'f':
		case 'm':
		case 'a':
		case 'p':
			p += 5;
			p += strlen(p)+1;
			continue;
		}
		val += ANOFFSET (objfile->section_offsets, section);
		prim_record_minimal_symbol_and_info(name, val, ms_type, NULL, section, bfd_section, objfile);
		p += 5;
		p += strlen(p)+1;
	}

	free(syms);

	if (pst)
	{
		/* Don't set pst->texthigh lower than it already is.  */
		CORE_ADDR text_end =
		(lowest_text_address == (CORE_ADDR) -1
		? (text_addr + ANOFFSET (objfile->section_offsets, SECT_OFF_TEXT))
		: lowest_text_address)
		+ text_size;

		plan9_end_psymtab (pst, psymtab_include_list, includes_used,
		   n,
		   text_end > pst->texthigh ? text_end : pst->texthigh,
		   dependency_list, dependencies_used, textlow_not_set);
	}

	do_cleanups (back_to);
}

/* Allocate and partially fill a partial symtab.  It will be
   completely filled at the end of the symbol list.

   SYMFILE_NAME is the name of the symbol-file we are reading from, and ADDR
   is the address relative to which its symbols are (incremental) or 0
   (normal). */


static struct partial_symtab *
plan9_start_psymtab (objfile, filename, textlow, ldsymoff, global_syms, static_syms)
     struct objfile *objfile;
     char *filename;
     CORE_ADDR textlow;
     int ldsymoff;
     struct partial_symbol **global_syms;
     struct partial_symbol **static_syms;
{
  struct partial_symtab *result =
  start_psymtab_common (objfile, objfile->section_offsets,
			filename, textlow, global_syms, static_syms);

  result->read_symtab_private = (char *)
    obstack_alloc (&objfile->psymbol_obstack, sizeof (struct symloc));
  LDSYMOFF (result) = ldsymoff;
  result->read_symtab = plan9_psymtab_to_symtab;
  SYMBOL_OFFSET (result) = symbol_table_offset;

  /* Deduce the source language from the filename for this psymtab. */
  psymtab_language = deduce_language_from_filename (filename);

  return result;
}

/* Close off the current usage of PST.  
   Returns PST or NULL if the partial symtab was empty and thrown away.

   FIXME:  List variables and peculiarities of same.  */

static struct partial_symtab *
plan9_end_psymtab (pst, include_list, num_includes, capping_symbol_offset,
	capping_text, dependency_list, number_dependencies, textlow_not_set)
     struct partial_symtab *pst;
     char **include_list;
     int num_includes;
     int capping_symbol_offset;
     CORE_ADDR capping_text;
     struct partial_symtab **dependency_list;
     int number_dependencies;
     int textlow_not_set;
{
  int i;
  struct objfile *objfile = pst->objfile;

  if (capping_symbol_offset != -1)
    LDSYMLEN (pst) = capping_symbol_offset - LDSYMOFF (pst);
  pst->texthigh = capping_text;

  pst->n_global_syms =
    objfile->global_psymbols.next - (objfile->global_psymbols.list + pst->globals_offset);
  pst->n_static_syms =
    objfile->static_psymbols.next - (objfile->static_psymbols.list + pst->statics_offset);

  pst->number_of_dependencies = number_dependencies;
  if (number_dependencies)
    {
      pst->dependencies = (struct partial_symtab **)
	obstack_alloc (&objfile->psymbol_obstack,
		    number_dependencies * sizeof (struct partial_symtab *));
      memcpy (pst->dependencies, dependency_list,
	      number_dependencies * sizeof (struct partial_symtab *));
    }
  else
    pst->dependencies = 0;

  for (i = 0; i < num_includes; i++)
    {
      struct partial_symtab *subpst =
      allocate_psymtab (include_list[i], objfile);

      subpst->section_offsets = pst->section_offsets;
      subpst->read_symtab_private =
	(char *) obstack_alloc (&objfile->psymbol_obstack,
				sizeof (struct symloc));
      LDSYMOFF (subpst) =
	LDSYMLEN (subpst) =
	subpst->textlow =
	subpst->texthigh = 0;

      /* We could save slight bits of space by only making one of these,
         shared by the entire set of include files.  FIXME-someday.  */
      subpst->dependencies = (struct partial_symtab **)
	obstack_alloc (&objfile->psymbol_obstack,
		       sizeof (struct partial_symtab *));
      subpst->dependencies[0] = pst;
      subpst->number_of_dependencies = 1;

      subpst->globals_offset =
	subpst->n_global_syms =
	subpst->statics_offset =
	subpst->n_static_syms = 0;

      subpst->readin = 0;
      subpst->symtab = 0;
      subpst->read_symtab = pst->read_symtab;
    }

  sort_pst_symbols (pst);

  /* If there is already a psymtab or symtab for a file of this name, remove it.
     (If there is a symtab, more drastic things also happen.)
     This happens in VxWorks.  */
  free_named_symtabs (pst->filename);

  if (num_includes == 0
      && number_dependencies == 0
      && pst->n_global_syms == 0
      && pst->n_static_syms == 0
      && has_line_numbers == 0)
    {
      /* Throw away this psymtab, it's empty.  We can't deallocate it, since
         it is on the obstack, but we can forget to chain it on the list.  */
      /* Empty psymtabs happen as a result of header files which don't have
         any symbols in them.  There can be a lot of them.  But this check
         is wrong, in that a psymtab with N_SLINE entries but nothing else
         is not empty, but we don't realize that.  Fixing that without slowing
         things down might be tricky.  */

      discard_psymtab (pst);

      /* Indicate that psymtab was thrown away.  */
      pst = (struct partial_symtab *) NULL;
    }
  return pst;
}

static void
plan9_psymtab_to_symtab_1 (pst)
     struct partial_symtab *pst;
{
  struct cleanup *old_chain;
  int i;

  if (!pst)
    return;

  if (pst->readin)
    {
      fprintf_unfiltered (gdb_stderr, "Psymtab for %s already read in.  Shouldn't happen.\n",
			  pst->filename);
      return;
    }

  /* Read in all partial symtabs on which this one is dependent */
  for (i = 0; i < pst->number_of_dependencies; i++)
    if (!pst->dependencies[i]->readin)
      {
	/* Inform about additional files that need to be read in.  */
	if (info_verbose)
	  {
	    fputs_filtered (" ", gdb_stdout);
	    wrap_here ("");
	    fputs_filtered ("and ", gdb_stdout);
	    wrap_here ("");
	    printf_filtered ("%s...", pst->dependencies[i]->filename);
	    wrap_here ("");	/* Flush output */
	    gdb_flush (gdb_stdout);
	  }
	plan9_psymtab_to_symtab_1 (pst->dependencies[i]);
      }

  if (LDSYMLEN (pst))		/* Otherwise it's a dummy */
    {
      /* Init stuff necessary for reading in symbols */
      stabsread_init ();
      buildsym_init ();
      old_chain = make_cleanup (really_free_pendings, 0);

      /* Read in this file's symbols */
      bfd_seek (pst->objfile->obfd, SYMBOL_OFFSET (pst), SEEK_SET);
      read_ofile_symtab (pst);
      sort_symtab_syms (pst->symtab);

      do_cleanups (old_chain);
    }

  pst->readin = 1;
}

/* Read in all of the symbols for a given psymtab for real.
   Be verbose about it if the user wants that.  */

static void
plan9_psymtab_to_symtab (pst)
     struct partial_symtab *pst;
{
  bfd *sym_bfd;

  if (!pst)
    return;

  if (pst->readin)
    {
      fprintf_unfiltered (gdb_stderr, "Psymtab for %s already read in.  Shouldn't happen.\n",
			  pst->filename);
      return;
    }

  if (LDSYMLEN (pst) || pst->number_of_dependencies)
    {
      /* Print the message now, before reading the string table,
         to avoid disconcerting pauses.  */
      if (info_verbose)
	{
	  printf_filtered ("Reading in symbols for %s...", pst->filename);
	  gdb_flush (gdb_stdout);
	}

      sym_bfd = pst->objfile->obfd;

      next_symbol_text_func = plan9_next_symbol_text;

      plan9_psymtab_to_symtab_1 (pst);

      /* Match with global symbols.  This only needs to be done once,
         after all of the symtabs and dependencies have been read in.   */
      scan_file_globals (pst->objfile);

      /* Finish up the debug error message.  */
      if (info_verbose)
	printf_filtered ("done.\n");
    }
}

/* Read in a defined section of a specific object file's symbols. */

static void
read_ofile_symtab (pst)
     struct partial_symtab *pst;
{
	register char *namestring;
	register struct external_nlist *bufp;
	struct internal_nlist nlist;
 	char *name, buf[5];
	unsigned char *syms, *p, *ep;
	CORE_ADDR val;
	int desc, index;
	unsigned char type;
	unsigned max_symnum;
	register bfd *abfd;
	struct objfile *objfile;
	int sym_offset;		/* Offset to start of symbols to read */
	int sym_size;			/* Size of symbols to read */
	CORE_ADDR text_offset;	/* Start of text segment for symbols */
	int text_size;		/* Size of text segment for symbols */
	struct section_offsets *section_offsets;

	objfile = pst->objfile;
	sym_offset = LDSYMOFF (pst);
	sym_size = LDSYMLEN (pst);
	text_offset = pst->textlow;
	text_size = pst->texthigh - pst->textlow;
	section_offsets = pst->section_offsets;

	current_objfile = objfile;
	subfile_stack = NULL;

	last_source_file = NULL;

	abfd = objfile->obfd;

	bfd_seek (abfd, sym_offset, SEEK_CUR);
	processing_gcc_compilation = 0;

	syms = (unsigned char *)bfd_malloc(sym_size);
	if (syms == NULL)
		perror_with_name (objfile->name);
	if ((int)bfd_read (syms, 1, sym_size, abfd) != sym_size) {
		free (syms);
		perror_with_name (objfile->name);
	}

 	p = syms;
	ep = p+sym_size;
	for(;;) {
		if(p >= ep)
			break;
		QUIT;			/* Allow this to be interruptable */
		OBJSTAT (objfile, n_stabs++);
		val = bfd_h_get_32(abfd, p);
		name = (char*)&p[5];
		switch(p[4] & ~0x80) {
		case 't':
			if (STREQ (namestring, GCC_COMPILED_FLAG_SYMBOL))
				processing_gcc_compilation = 1;
			else if (STREQ (namestring, GCC2_COMPILED_FLAG_SYMBOL))
				processing_gcc_compilation = 2;

			if (AUTO_DEMANGLING)
				set_demangling_style (GNU_DEMANGLING_STYLE_STRING);
			break;
		case 'X':			/* encapsulated stabs */
			buf[0] = name[0];
			buf[1] = name[1];
			buf[2] = '\0';
			type = strtoul(buf, (char **)NULL, 16);
			buf[0] = name[4];
			buf[1] = name[5];
			buf[2] = name[6];
			buf[3] = name[7];
			buf[4] = '\0';
			desc = strtoul(buf, (char **)NULL, 16);

			if (type & N_STAB)
				plan9_process_one_symbol (type, desc, val, name+8, section_offsets, objfile);
			break;
		case 'z':
			p += 5+1;
			for(;;) {
				if(p+1 >= ep)
					break;
				index = (p[0]<<8) | p[1];
				p += 2;
				if(index == 0)
					break;
			}
			continue;
		}
		p += 5;
		p += strlen(p)+1;
	}

	free(syms);

	current_objfile = NULL;

	/* In reordered executables last_source_start_addr may not be the
	lower bound for this symtab, instead use text_offset which comes
	from pst->textlow which is correct.  */
	if (last_source_start_addr > text_offset)
		last_source_start_addr = text_offset;

	pst->symtab = end_symtab (text_offset + text_size, objfile, SECT_OFF_TEXT);

	end_stabs ();
}


/* This handles a single symbol from the symbol-file, building symbols
   into a GDB symtab.  It takes these arguments and an implicit argument.

   TYPE is the type field of the ".stab" symbol entry.
   DESC is the desc field of the ".stab" entry.
   VALU is the value field of the ".stab" entry.
   NAME is the symbol name, in our address space.
   SECTION_OFFSETS is a set of amounts by which the sections of this object
   file were relocated when it was loaded into memory.
   All symbols that refer
   to memory locations need to be offset by these amounts.
   OBJFILE is the object file from which we are reading symbols.
   It is used in end_symtab.  */

static void
plan9_process_one_symbol (type, desc, valu, name, section_offsets, objfile)
     int type, desc;
     CORE_ADDR valu;
     char *name;
     struct section_offsets *section_offsets;
     struct objfile *objfile;
{
  register struct context_stack *new;
  /* This remembers the address of the start of a function.  It is used
     because in Solaris 2, N_LBRAC, N_RBRAC, and N_SLINE entries are
     relative to the current function's start address.  On systems
     other than Solaris 2, this just holds the SECT_OFF_TEXT value, and is
     used to relocate these symbol types rather than SECTION_OFFSETS.  */
  static CORE_ADDR function_start_offset;

  /* If this is nonzero, we've seen a non-gcc N_OPT symbol for this source
     file.  Used to detect the SunPRO solaris compiler.  */
  static int n_opt_found;

  /* The stab type used for the definition of the last function.
     N_STSYM or N_GSYM for SunOS4 acc; N_FUN for other compilers.  */
  static int function_stab_type = 0;

    /* N_LBRAC, N_RBRAC and N_SLINE entries are not relative to the
       function start address, so just use the text offset.  */
    function_start_offset = ANOFFSET (section_offsets, SECT_OFF_TEXT);

  /* Something is wrong if we see real data before
     seeing a source file name.  */

  if (last_source_file == NULL && type != (unsigned char) N_SO)
    {
      /* Ignore any symbols which appear before an N_SO symbol.
         Currently no one puts symbols there, but we should deal
         gracefully with the case.  A complain()t might be in order,
         but this should not be an error ().  */
      return;
    }

  switch (type)
    {
    case N_FUN:
    case N_FNAME:

      if (*name == '\000')
	{
	  /* This N_FUN marks the end of a function.  This closes off the
	     current block.  */
	  within_function = 0;
	  new = pop_context ();

	  /* Make a block for the local symbols within.  */
	  finish_block (new->name, &local_symbols, new->old_blocks,
			new->start_addr, new->start_addr + valu,
			objfile);

	  break;
	}

      /* Relocate for dynamic loading */
      valu += ANOFFSET (section_offsets, SECT_OFF_TEXT);
#ifdef SMASH_TEXT_ADDRESS
      SMASH_TEXT_ADDRESS (valu);
#endif
      goto define_a_symbol;

    case N_LBRAC:
      /* This "symbol" just indicates the start of an inner lexical
         context within a function.  */

      /* Ignore extra outermost context from SunPRO cc and acc.  */
      if (n_opt_found && desc == 1)
	break;

#ifdef SUN_FIXED_LBRAC_BUG
      if (!SUN_FIXED_LBRAC_BUG && valu < last_pc_address)
	{
	  /* Patch current LBRAC pc value to match last handy pc value */
	  complain (&lbrac_complaint);
	  valu = last_pc_address;
	}
#endif
      new = push_context (desc, valu);
      break;

    case N_RBRAC:
      /* This "symbol" just indicates the end of an inner lexical
         context that was started with N_LBRAC.  */

      /* Ignore extra outermost context from SunPRO cc and acc.  */
      if (n_opt_found && desc == 1)
	break;

      new = pop_context ();
      if (desc != new->depth)
	complain (&lbrac_mismatch_complaint, symnum);

      /* Some compilers put the variable decls inside of an
         LBRAC/RBRAC block.  This macro should be nonzero if this
         is true.  DESC is N_DESC from the N_RBRAC symbol.
         GCC_P is true if we've detected the GCC_COMPILED_SYMBOL
         or the GCC2_COMPILED_SYMBOL.  */
#if !defined (VARIABLES_INSIDE_BLOCK)
#define VARIABLES_INSIDE_BLOCK(desc, gcc_p) 0
#endif

      /* Can only use new->locals as local symbols here if we're in
         gcc or on a machine that puts them before the lbrack.  */
      if (!VARIABLES_INSIDE_BLOCK (desc, processing_gcc_compilation))
	local_symbols = new->locals;

      if (context_stack_depth
	  > !VARIABLES_INSIDE_BLOCK (desc, processing_gcc_compilation))
	{
	  /* This is not the outermost LBRAC...RBRAC pair in the function,
	     its local symbols preceded it, and are the ones just recovered
	     from the context stack.  Define the block for them (but don't
	     bother if the block contains no symbols.  Should we complain
	     on blocks without symbols?  I can't think of any useful purpose
	     for them).  */
	  if (local_symbols != NULL)
	    {
	      /* Muzzle a compiler bug that makes end < start.  (which
	         compilers?  Is this ever harmful?).  */
	      if (new->start_addr > valu)
		{
		  complain (&lbrac_rbrac_complaint);
		  new->start_addr = valu;
		}
	      /* Make a block for the local symbols within.  */
	      finish_block (0, &local_symbols, new->old_blocks,
			    new->start_addr, valu, objfile);
	    }
	}
      else
	{
	  /* This is the outermost LBRAC...RBRAC pair.  There is no
	     need to do anything; leave the symbols that preceded it
	     to be attached to the function's own block.  We need to
	     indicate that we just moved outside of the function.  */
	  within_function = 0;
	}

      if (VARIABLES_INSIDE_BLOCK (desc, processing_gcc_compilation))
	/* Now pop locals of block just finished.  */
	local_symbols = new->locals;
      break;

    case N_FN:
    case N_FN_SEQ:
      /* This kind of symbol indicates the start of an object file.  */
      /* Relocate for dynamic loading */
      valu += ANOFFSET (section_offsets, SECT_OFF_TEXT);
      break;

    case N_SO:
      /* This type of symbol indicates the start of data
         for one source file.
         Finish the symbol table of the previous source file
         (if any) and start accumulating a new symbol table.  */
      /* Relocate for dynamic loading */
      valu += ANOFFSET (section_offsets, SECT_OFF_TEXT);

      n_opt_found = 0;

#ifdef SUN_FIXED_LBRAC_BUG
      last_pc_address = valu;	/* Save for SunOS bug circumcision */
#endif

#ifdef PCC_SOL_BROKEN
      /* pcc bug, occasionally puts out SO for SOL.  */
      if (context_stack_depth > 0)
	{
	  start_subfile (name, NULL);
	  break;
	}
#endif
      if (last_source_file)
	{
	  /* Check if previous symbol was also an N_SO (with some
	     sanity checks).  If so, that one was actually the directory
	     name, and the current one is the real file name.
	     Patch things up. */
	  if (previous_stab_code == (unsigned char) N_SO)
	    {
	      patch_subfile_names (current_subfile, name);
	      break;		/* Ignore repeated SOs */
	    }
	  end_symtab (valu, objfile, SECT_OFF_TEXT);
	  end_stabs ();
	}

      /* Null name means this just marks the end of text for this .o file.
         Don't start a new symtab in this case.  */
      if (*name == '\000')
	break;

      start_stabs ();
      start_symtab (name, NULL, valu);
      record_debugformat ("stabs");
      break;

    case N_SOL:
      /* This type of symbol indicates the start of data for
         a sub-source-file, one whose contents were copied or
         included in the compilation of the main source file
         (whose name was given in the N_SO symbol.)  */
      /* Relocate for dynamic loading */
      valu += ANOFFSET (section_offsets, SECT_OFF_TEXT);
      start_subfile (name, current_subfile->dirname);
      break;

    case N_BINCL:
      push_subfile ();
      add_new_header_file (name, valu);
      start_subfile (name, current_subfile->dirname);
      break;

    case N_EINCL:
      start_subfile (pop_subfile (), current_subfile->dirname);
      break;

    case N_EXCL:
      add_old_header_file (name, valu);
      break;

    case N_SLINE:
      /* This type of "symbol" really just records
         one line-number -- core-address correspondence.
         Enter it in the line list for this symbol table.  */

      /* Relocate for dynamic loading and for ELF acc fn-relative syms.  */
      valu += function_start_offset;

#ifdef SUN_FIXED_LBRAC_BUG
      last_pc_address = valu;	/* Save for SunOS bug circumcision */
#endif
      record_line (current_subfile, desc, valu);
      break;

    case N_BCOMM:
      common_block_start (name, objfile);
      break;

    case N_ECOMM:
      common_block_end (objfile);
      break;

      /* The following symbol types need to have the appropriate offset added
         to their value; then we process symbol definitions in the name.  */

    case N_STSYM:		/* Static symbol in data seg */
    case N_DSLINE:		/* Source line number, data seg */
      valu += ANOFFSET (section_offsets, SECT_OFF_DATA);
      goto define_a_symbol;

    case N_LCSYM:		/* Static symbol in BSS seg */
    case N_BSLINE:		/* Source line number, bss seg */
      /*   N_BROWS:       overlaps with N_BSLINE */
      valu += ANOFFSET (section_offsets, SECT_OFF_BSS);
      goto define_a_symbol;

    case N_ROSYM:		/* Static symbol in Read-only data seg */
      valu += ANOFFSET (section_offsets, SECT_OFF_RODATA);
      goto define_a_symbol;

    case N_ENTRY:		/* Alternate entry point */
      /* Relocate for dynamic loading */
      valu += ANOFFSET (section_offsets, SECT_OFF_TEXT);
      goto define_a_symbol;

      /* The following symbol types we don't know how to process.  Handle
         them in a "default" way, but complain to people who care.  */
    default:
    case N_CATCH:		/* Exception handler catcher */
    case N_EHDECL:		/* Exception handler name */
    case N_PC:			/* Global symbol in Pascal */
    case N_M2C:		/* Modula-2 compilation unit */
      /*   N_MOD2:        overlaps with N_EHDECL */
    case N_SCOPE:		/* Modula-2 scope information */
    case N_ECOML:		/* End common (local name) */
    case N_NBTEXT:		/* Gould Non-Base-Register symbols??? */
    case N_NBDATA:
    case N_NBBSS:
    case N_NBSTS:
    case N_NBLCS:
      complain (&unknown_symtype_complaint, local_hex_string (type));
      /* FALLTHROUGH */

      /* The following symbol types don't need the address field relocated,
         since it is either unused, or is absolute.  */
    define_a_symbol:
    case N_GSYM:		/* Global variable */
    case N_NSYMS:		/* Number of symbols (ultrix) */
    case N_NOMAP:		/* No map?  (ultrix) */
    case N_RSYM:		/* Register variable */
    case N_DEFD:		/* Modula-2 GNU module dependency */
    case N_SSYM:		/* Struct or union element */
    case N_LSYM:		/* Local symbol in stack */
    case N_PSYM:		/* Parameter variable */
    case N_LENG:		/* Length of preceding symbol type */
      if (name)
	{
	  int deftype;
	  char *colon_pos = strchr (name, ':');
	  if (colon_pos == NULL)
	    deftype = '\0';
	  else
	    deftype = colon_pos[1];

	  switch (deftype)
	    {
	    case 'f':
	    case 'F':
	      function_stab_type = type;
	      within_function = 1;

	      if (context_stack_depth > 1)
		{
		  complain (&lbrac_unmatched_complaint, symnum);
		  break;
		}

	      if (context_stack_depth > 0)
		{
		  new = pop_context ();
		  /* Make a block for the local symbols within.  */
		  finish_block (new->name, &local_symbols, new->old_blocks,
				new->start_addr, valu, objfile);
		}

	      new = push_context (0, valu);
	      new->name = define_symbol (valu, name, desc, type, objfile);
	      break;

	    default:
	      define_symbol (valu, name, desc, type, objfile);
	      break;
	    }
	}
      break;

      /* We use N_OPT to carry the gcc2_compiled flag.  Sun uses it
         for a bunch of other flags, too.  Someday we may parse their
         flags; for now we ignore theirs and hope they'll ignore ours.  */
    case N_OPT:		/* Solaris 2:  Compiler options */
      if (name)
	{
	  if (STREQ (name, GCC2_COMPILED_FLAG_SYMBOL))
	    {
	      processing_gcc_compilation = 2;
#if 1				/* Works, but is experimental.  -fnf */
	      if (AUTO_DEMANGLING)
		{
		  set_demangling_style (GNU_DEMANGLING_STYLE_STRING);
		}
#endif
	    }
	  else
	    n_opt_found = 1;
	}
      break;

      /* The following symbol types can be ignored.  */
    case N_OBJ:		/* Solaris 2:  Object file dir and name */
      /*   N_UNDF:                   Solaris 2:  file separator mark */
      /*   N_UNDF: -- we will never encounter it, since we only process one
         file's symbols at once.  */
    case N_ENDM:		/* Solaris 2:  End of module */
    case N_MAIN:		/* Name of main routine.  */
    case N_ALIAS:		/* SunPro F77: alias name, ignore for now.  */
      break;
    }

  /* '#' is a GNU C extension to allow one symbol to refer to another
     related symbol.

     Generally this is used so that an alias can refer to its main
     symbol.  */
  if (name[0] == '#')
    {
      /* Initialize symbol reference names and determine if this is 
         a definition.  If symbol reference is being defined, go 
         ahead and add it.  Otherwise, just return sym. */

      char *s = name;
      int refnum;

      /* If this stab defines a new reference ID that is not on the
         reference list, then put it on the reference list.

         We go ahead and advance NAME past the reference, even though
         it is not strictly necessary at this time.  */
      refnum = symbol_reference_defined (&s);
      if (refnum >= 0)
	if (!ref_search (refnum))
	  ref_add (refnum, 0, name, valu);
      name = s;
    }


  previous_stab_code = type;
}

static struct sym_fns plan9_sym_fns =
{
  bfd_target_plan9_flavour,
  plan9_new_init,			/* sym_new_init: init anything gbl to entire symtab */
  plan9_symfile_init,		/* sym_init: read initial info, setup for sym_read() */
  plan9_symfile_read,		/* sym_read: read a symbol file into symtab */
  plan9_symfile_finish,		/* sym_finish: finished with file, cleanup */
  default_symfile_offsets,	/* sym_offsets: parse user's offsets to internal form */
  NULL				/* next: pointer to next struct sym_fns */
};

void
_initialize_plan9read ()
{
  add_symtab_fns (&plan9_sym_fns);
}
